// Space-Traversal-Matching

// Created by Sander Huisman and Mickael Bourgoin on 10/05/2017.

// Modified by Sander Huisman on 02/03/2020:
//    Removed the CMatrix class, inverse/determinant of 3*3 matrix hard-coded => 33% faster.

// Modified by Sander Huisman in 21/11/2020:
//    Const reference passing arguments and emplace_back in some places => 35% faster.

// Modified by Sander Huisman in 25/2020:
//    Changed some algorithms => 30% faster.

// Visit http://shuisman.com/index.php?STM for some details.


#include <algorithm>
#include <ctime>
#include <iostream>
#include <fstream>
#include <vector>
#include <math.h>
#include <map>
#include <tuple>

double timing1, timing2, timing3, timing4, timing5, timing6, timing7, timing8, timing9, timing10 = 0; // Timings that are used, all the lines with txb and txe can be removed
std::clock_t t1b, t1e = clock();
std::clock_t t2b, t2e = clock();
std::clock_t t3b, t3e = clock();
std::clock_t t4b, t4e = clock();
std::clock_t t5b, t5e = clock();
std::clock_t t6b, t6e = clock();
std::clock_t t7b, t7e = clock();
std::clock_t t8b, t8e = clock();
std::clock_t t9b, t9e = clock();
std::clock_t t10b, t10e = clock();

std::ofstream logstream;    // Stream for logging output

std::vector<std::vector<int>> neighbours;   // Stores the relative indices of neighbours
std::vector<std::vector<double>> bounds;    // Stores the bounds of the voxels (for each direction x,y,z this is a vector with all the bounds)


// Various structs that are used to store data
struct cellvisit{
    cellvisit(double time_, int dimension_, bool edgeQ_): time(time_), dimension(dimension_), edgeQ(edgeQ_){}
    cellvisit(){}
    double time;
    int dimension = -1;
    bool edgeQ = false;
};

struct cellid{
    long xi;
    long yi;
    long zi;
};

struct traversedcell{
    int camid = -1;
    int rayid = -1;
    struct cellid cellid;
};

struct camrayid{
    long camid;
    long rayid;
};

struct candidatematch{
    double matchx = 0;
    double matchy = 0;
    double matchz = 0;
    double matcherror = -1;
    std::vector<camrayid> camrayids;
};

struct groupedcell{
    struct cellid cellid;
    std::vector<camrayid> camrayids;
};

struct ray{
    int camid = -1;
    int rayid = -1;
    double x;
    double y;
    double z;
    double vx;
    double vy;
    double vz;
    double tstart;
    double tend;
};

struct transformedray{
    int camid = -1;
    int rayid = -1;
    double x;
    double y;
    double z;
    double vx;
    double vy;
    double vz;
    double tstart;
    double tend;
    bool hit = false;
    bool inside = false;
};

struct boundingboxspec{
    double xmin;
    double xmax;
    double ymin;
    double ymax;
    double zmin;
    double zmax;
    int nx;
    int ny;
    int nz;
};

struct hitpoint{
    double t;
    double posx;
    double posy;
    double posz;
};


// Sorting functions for custom structs(classes). These tell how the above structs are considered sorted.
bool comparecellvisitbytime(const cellvisit &a, const cellvisit &b)
{
    return a.time < b.time;
}

bool comparecandidatematches(const candidatematch &a, const candidatematch &b) // First sort by number of cams then by match error
{
    return(a.camrayids.size() > b.camrayids.size() || (a.camrayids.size() == b.camrayids.size() && a.matcherror < b.matcherror));
}

bool comparecamrayidscam(const camrayid &a, const camrayid &b)
{
    return(a.camid < b.camid);
}

bool comparecamrayididentical(const camrayid &a, const camrayid &b)
{
    return(a.camid == b.camid && a.rayid == b.rayid);
}

bool comparecamrayidscamray(const camrayid &a, const camrayid &b)
{
    return(a.camid < b.camid || (a.camid == b.camid && a.rayid < b.rayid));
}

bool comparecamrayidsamecam(const camrayid &a, const camrayid &b)
{
    return(a.camid == b.camid);
}

bool comparecamrayidsidentical(const std::vector<camrayid> &a, const std::vector<camrayid> &b)
{
    if(a.size() != b.size())
    {
        return(false);
    }
    else
    {
        for(int i = 0; i < a.size();i++)
        {
            if(a[i].camid != b[i].camid || a[i].rayid != b[i].rayid)
            {
                return(false);
            }
        }
        return(true);
    }
}

bool comparecamrayidsordered(const std::vector<camrayid> &a, const std::vector<camrayid> &b)
{
    size_t as = a.size();
    size_t bs = b.size();
    long minlen = std::min(as,bs);
    for(size_t i=0; i< minlen; i++)
    {
        if(a[i].camid < b[i].camid || (a[i].camid == b[i].camid && a[i].rayid < b[i].rayid))
        {
            return(true);
        }
        else
        {
            if(a[i].camid > b[i].camid || (a[i].camid == b[i].camid && a[i].rayid > b[i].rayid))
            {
                return(false);
            }
        }
    }
    if(as != bs)
    {
        return(as < bs);    // Short goes before long
    }
    else                    // This basically means that they have same length and content
    {
        return(false);
    }
}

bool comparetraversedcell(const traversedcell &a, const traversedcell &b)
{
    return(a.cellid.xi < b.cellid.xi || (a.cellid.xi == b.cellid.xi && a.cellid.yi < b.cellid.yi) || (a.cellid.xi == b.cellid.xi && a.cellid.yi == b.cellid.yi && a.cellid.zi < b.cellid.zi));
}

bool comparetraversedcellsamecell(const traversedcell &a, const traversedcell &b)
{
    return(a.cellid.xi == b.cellid.xi && a.cellid.yi == b.cellid.yi && a.cellid.zi == b.cellid.zi);
}

bool comparehitpointbytime(const hitpoint &a, const hitpoint &b)
{
    return a.t < b.t;
}


// Some mathematical functions
template <typename number> // I don't check explicitly if it is a number, but it is presumed to be a number (int, long, single, double...)
number specialdivision(number a,number b){      // Captures division by 0 in the very rare case the rays are exactly along a certain direction. In the case of b=0 the answer will be -infinity.
    if(b==0.0){
        return(-1*INFINITY);
    }
    else{
        return(a/b);
    }
}

template <typename number>
number squarenorm(number vector[3]) {       // Squared Euclidean norm
    return(vector[0]*vector[0] + vector[1]*vector[1] + vector[2]*vector[2]);
}

template <typename number>
number norm(number vector[3]) {             // Euclidean norm
    return(sqrt(squarenorm(vector)));
}

template <typename number>                  // Helper function to check if bmin <= b <= bmax
bool atface(number bmin,number bmax,number b){
    return(bmin <=b && b <= bmax);
}

template <typename number>
signed int Sgn(number x)                    // Modified sign function where the case of 0 returns -1 rather than 0.
{
    if (x<0)
        return(-1);
    else
        return(1);
}

long positionsorted(const std::vector<double>& boundaries, const double& n) // Find in O(Log(n)) time the position of a number inside a sorted list…
{
    long mn = 0;
    long mx = boundaries.size() - 1;
    long trial;
    if(boundaries[mn]-1.0e-8 <= n && n <= boundaries[mx]+1.0e-8)
    {
        while(mx-mn > 1)
        {
            trial = (mn + mx)>>1;
            if(n > boundaries[trial])
            {
                mn = trial;
            }
            else
            {
                mx = trial;
            }
        }
        return(mn);
    }
    else
    {
        return(-1);
    }
}

// Recursive function to find the permutation of a list of lists (all tuples) // Second argument is modified so can not be passed as const.
void GenerateCamRayIDPermutations(const std::vector<std::vector<camrayid>>& Lists, std::vector<std::vector<camrayid>>& result, const int& depth, const std::vector<camrayid>& current)
{
    if(depth == Lists.size())
    {
        result.push_back(current);
        return;
    }
    
    for(int i = 0; i < Lists[depth].size(); ++i)
    {
        std::vector<camrayid> tmp;
        tmp = current;
        tmp.push_back(Lists[depth][i]);
        GenerateCamRayIDPermutations(Lists, result, depth + 1, tmp);
    }
}

// Taylor-Couette specific code: do not traverse to voxels that are fully outside the gap.
// Could be modified for pipe flow to make sure it does not match outside the pipe et cetera.
bool FullyOutsideGap(std::vector<std::vector<double>> bounds, cellid cids, double ri, double ro)
{
    double risq = ri*ri;
    double rosq = ro*ro;
    
    double xmin = bounds[0][cids.xi];
    double xmax = bounds[0][cids.xi+1];
    double ymin = bounds[1][cids.yi];
    double ymax = bounds[1][cids.yi+1];
    
    double rsq;
    double xminsq = xmin*xmin;
    double yminsq = ymin*ymin;
    rsq = xminsq + yminsq; // Check if all 4 corners of the cells are outside the gap (assuming that a cell does not cross the entire gap).
    if(rsq<risq || rsq > rosq)
    {
        double ymaxsq = ymax*ymax;
        rsq = xminsq + ymaxsq;
        if(rsq<risq || rsq > rosq)
        {
            double xmaxsq = xmax*xmax;
            rsq = xmaxsq + yminsq;
            if(rsq<risq || rsq > rosq)
            {
                rsq = xmaxsq + ymaxsq;
                if(rsq<risq || rsq > rosq)
                {
                    return true; // Inside
                }
            }
        }
    }
    return false; // Outside
}

// Merges two sorted lists into a single sorted list
std::vector<cellvisit> MergeSortedLists(const std::vector<cellvisit>& cv1, const std::vector<cellvisit>& cv2)
{
    unsigned long len1 = cv1.size();
    //std::cout << "len cv1 = " << len1 << std::endl;
    unsigned long len2 = cv2.size();
    //std::cout << "len cv2 = " << len2 << std::endl;
    if(len1 > 0 && len2 > 0)
    {
        unsigned long len = len1 + len2;
        //std::cout << "len = " << len << std::endl;
        std::vector<cellvisit> out(len);

        unsigned long cp1 = 0;
        unsigned long cp2 = 0;
        for (int i=0; i < len; i++) {
            //std::cout << " i = " << i << "/" << len << std::endl;
            if(cv1[cp1].time < cv2[cp2].time)
            {
                out[i] = cv1[cp1];
                if(cp1<len1)
                {
                    cp1++;
                }
                else
                {
                    for(int j = i; j < len; j++)
                    {
                        out[j] = cv2[cp2];
                        cp2++;
                    }
                    break;
                }
            }
            else
            {
                out[i] = cv2[cp2];
                if(cp2<len2)
                {
                    cp2++;
                }
                else
                {
                    for(int j = i; j < len; j++)
                    {
                        out[j] = cv1[cp1];
                        cp1++;
                    }
                    break;
                }
            }
        }
        return(out);
    }
    else
    {
        if(len1 == 0)
        {
            return(cv2);
        }
        else
        {
            return(cv1);
        }
    }

}

// Traverse a single ray in a single direction and find the voxels it traverses.
std::vector<traversedcell> DirectionalVoxelTraversal(const transformedray& ray, const std::vector<std::vector<double>>& bounds, const double& tmax)
{
    std::vector<traversedcell> out;
    traversedcell tmptraversedcell;
    tmptraversedcell.camid = ray.camid;
    tmptraversedcell.rayid = ray.rayid;
    cellid curindex;
    curindex.xi = positionsorted(bounds[0], ray.x);
    curindex.yi = positionsorted(bounds[1], ray.y);
    curindex.zi = positionsorted(bounds[2], ray.z);
    
//    if(ray.camid == 1 && ray.rayid == 50)
//    {
//        std::cout << std::endl << " raypos: = " << ray.x << " " << ray.y << " " << ray.z << std::endl;
//        std::cout << std::endl << "curindex = " << curindex.xi << " " << curindex.yi << " " << curindex.zi << std::endl;
//    }
    
    if(curindex.xi == -1 || curindex.yi == -1 || curindex.zi == -1) // Should not happen because before we project it onto the bounding box.
    {
        std::cout << "ray starts outside bounds!" << std::endl;
        logstream << "ray starts outside bounds!" << std::endl;
        std::cout << "p:" << ray.x << ", " << ray.y <<  ", " << ray.z << std::endl;
        logstream << "p:" << ray.x << ", " << ray.y <<  ", " << ray.z << std::endl;
        std::cout << "v:" << ray.vx << ", " << ray.vy <<  ", " << ray.vz << std::endl;
        logstream << "v:" << ray.vx << ", " << ray.vy <<  ", " << ray.vz << std::endl;
        std::cout << "v:" << curindex.xi << ", " << curindex.yi  <<  ", " << curindex.zi  << std::endl;
        logstream << "v:" << curindex.xi << ", " << curindex.yi  <<  ", " << curindex.zi  << std::endl;
        return(out);
    }
    else
    {
        long direction[3];
        direction[0] = Sgn(ray.vx);
        direction[1] = Sgn(ray.vy);
        direction[2] = Sgn(ray.vz);

        //traverse x, y, an z directions
        double tmptime;
        bool tmpedge;
        std::vector<cellvisit> cellvisits;
        cellvisits.reserve(bounds[0].size()); // Reverse max size already
        std::vector<cellvisit> cellvisitspart2;
        cellvisitspart2.reserve(bounds[1].size()); // Reverse max size already
        std::vector<cellvisit> cellvisitspart3;
        cellvisitspart3.reserve(bounds[2].size()); // Reverse max size already
        
        for(int i = 0; i<3; i++)        //For each dimension
        {
            double b = 0;
            double v = 0;
            switch (i) {
                case 0:
                    b = ray.x;
                    v = ray.vx;
                    break;
                case 1:
                    b = ray.y;
                    v = ray.vy;
                    break;
                case 2:
                    b = ray.z;
                    v = ray.vz;
            }
            
            for(int j = 0; j<bounds[i].size(); j++)
            {
                tmptime = specialdivision(bounds[i][j]-b,v);
                if(tmptime>1e-10)     // Only add those for which t is positive.  'ray is moving forward'
                {
                    tmpedge = false;
                    if(direction[i] == 1 && j == bounds[i].size() -1)
                    {
                        tmpedge = true;
                    }
                    else if(direction[i] == -1 && j == 0)
                    {
                        tmpedge = true;
                    }
                    if(i==0)
                    {
                        cellvisits.emplace_back(tmptime, i, tmpedge);
                    }
                    else if(i==1)
                    {
                        cellvisitspart2.emplace_back(tmptime, i, tmpedge);
                    }
                    else
                    {
                        cellvisitspart3.emplace_back(tmptime, i, tmpedge);
                    }
                }
            }
        }
 
        // Reverse if time is decreasing
        if(cellvisits[0].time > cellvisits[cellvisits.size()-1].time)
        {
            std::reverse(std::begin(cellvisits), std::end(cellvisits));
        }
        if(cellvisitspart2[0].time > cellvisitspart2[cellvisitspart2.size()-1].time)
        {
            std::reverse(std::begin(cellvisitspart2), std::end(cellvisitspart2));
        }
        if(cellvisitspart3[0].time > cellvisitspart3[cellvisitspart3.size()-1].time)
        {
            std::reverse(std::begin(cellvisitspart3), std::end(cellvisitspart3));
        }
        
        // Merge sorted lists into new sorted list:
        cellvisits = MergeSortedLists(cellvisits,cellvisitspart2);
        cellvisits = MergeSortedLists(cellvisits,cellvisitspart3);
        
        // Prevent from traversing outside bounding box
        for(long i = 0; cellvisits.size()-1;i++)
        {
            if(cellvisits[i].edgeQ == true)
            {
//                std::cout << "len = " << i << std::endl;
                cellvisits.resize(i); // Cut list short
                break;
            }
        }
        
        //std::cout << "len3-" << cellvisits.size() << std::endl;
        
//        if(ray.camid == 1 && ray.rayid == 50)
//        {
//            std::cout << "Postlength: " << cellvisits.size() << " // ";
//            std::cout << "curindex " << curindex.xi << "."  << curindex.yi << "."  << curindex.zi << std::endl;
//            for(auto cv:cellvisits)
//            {
//                std::cout << cv.dimension << " " <<cv.time << " " << cv.edgeQ << std::endl;
//            }
//        }
        
//        double ri = 0.075;
//        double ro = 0.105;
        
        // Reconstruct the traversed cells
        tmptraversedcell.cellid = curindex;                  // Start at curindex
        out.reserve(cellvisits.size());
        out.push_back(tmptraversedcell);
        for (const cellvisit& i: cellvisits)                // For each arrival time jump to next one...
        {
            
            switch (i.dimension) {
                case 0:
                    tmptraversedcell.cellid.xi += direction[i.dimension];
                    break;
                case 1:
                    tmptraversedcell.cellid.yi += direction[i.dimension];
                    break;
                case 2:
                    tmptraversedcell.cellid.zi += direction[i.dimension];
                    break;
                default:
                    break;
            }
//            Taylor-Couette mode: prevent from going inside IC, and outside OC: stop traversing once fully outside the gap
//            if(FullyOutsideGap(bounds, tmptraversedcell.cellid, ri, ro)) {
//                break;
//            }
         
            if (i.time > tmax) {
//                std::cout << "That took you long enough, mate!" << std::endl;
//                logstream << "That took you long enough, mate!" << std::endl;
//                std::cout << "T_begin, T_end: " << ray.tb << ", " << ray.te << std::endl;
//                logstream << "T_begin, T_end: " << ray.tb << ", " << ray.te << std::endl;
                break;
            }
            
            out.push_back(tmptraversedcell);
        }
        
//        if(ray.camid == 1 && ray.rayid == 50)
//        {
//            std::cout << std::endl << std::endl;
//            for(auto i: out)
//            {
//                std::cout << i.cellid.xi << " " << i.cellid.yi << " " << i.cellid.zi << std::endl;
//            }
//            std::cout << std::endl << std::endl;
//        }
        
        // Use neighbours to 'expand out' out, and remove duplicates.
        long len = out.size();
        out.resize(len*(neighbours.size()+1)); // Resize memory  +1 because of the center cells themselves
        long curpos = len;
        for(long i = 0; i <len; i++)
        {
            for(const std::vector<int>& j: neighbours) // Center cells are 0--size-1 in out vector. neighbours are 'appended' to that
            {
                out[curpos].camid = out[i].camid;
                out[curpos].rayid = out[i].rayid;
                out[curpos].cellid.xi = out[i].cellid.xi + j[0];
                out[curpos].cellid.yi = out[i].cellid.yi + j[1];
                out[curpos].cellid.zi = out[i].cellid.zi + j[2];
                curpos++;
            }
        }
        
        
        // Sort + delete duplicates
        t9b = clock();
        std::sort(out.begin(),out.end(),comparetraversedcell);
        t9e = clock();
        timing9 += double(t9e-t9b);
        
        out.erase(unique(out.begin(), out.end(), comparetraversedcellsamecell), out.end());
        out.shrink_to_fit();

        return(out);
    }
}

// Prepare input ray: Normalized the direction v to be 1, and modifies tstart and tend accordingly and applies the delta.
// If it starts outside the bounding box it projects the ray on to the bounding box.
// If it starts outside the bounding box and misses the bounding box, it is marked as such.
transformedray PrepareRay(const ray& r, const boundingboxspec& bb, const double& deltat)
{
    transformedray out;
    out.camid = r.camid;
    out.rayid = r.rayid;
    out.x = r.x;
    out.y = r.y;
    out.z = r.z;
    double nrm = sqrt(r.vx*r.vx + r.vy*r.vy + r.vz*r.vz);
    out.vx = r.vx / nrm;
    out.vy = r.vy / nrm;
    out.vz = r.vz / nrm;
    out.tstart = r.tstart * nrm - deltat;
    out.tend = r.tend * nrm + deltat;

    if(bb.xmin < out.x && out.x < bb.xmax && bb.ymin < out.y && out.y < bb.ymax && bb.zmin < out.z && out.z < bb.zmax) // Check if it is inside
    {
        out.inside = true;
        out.hit = true;
        return out;
    }
    else // Outside; but does it hit the bounding box
    {
        double t[6];
        t[0] = specialdivision(bb.xmin - out.x,out.vx);         // Find time until it hits the 6 planes (2 for each dimension)
        t[1] = specialdivision(bb.xmax - out.x,out.vx);
        t[2] = specialdivision(bb.ymin - out.y,out.vy);
        t[3] = specialdivision(bb.ymax - out.y,out.vy);
        t[4] = specialdivision(bb.zmin - out.z,out.vz);
        t[5] = specialdivision(bb.zmax - out.z,out.vz);
        
        std::vector<double> tmpip;
        std::vector<std::vector<double>> ip;
        double ti;
        for(int i=0; i<6; i++)
        {
            tmpip.clear();
            ti = t[i];
            if(ti == INFINITY || ti == -1*INFINITY)
            {
                tmpip.push_back(INFINITY);
                tmpip.push_back(INFINITY);
                tmpip.push_back(INFINITY);
            }
            else{
                tmpip.push_back(out.x + out.vx*ti);
                tmpip.push_back(out.y + out.vy*ti);
                tmpip.push_back(out.z + out.vz*ti);
            }
            ip.push_back(tmpip);
        }
        
        std::vector<bool> atfaces;
        atfaces.push_back(atface(bb.ymin, bb.ymax, ip[0][1]) && atface(bb.zmin, bb.zmax, ip[0][2]));
        atfaces.push_back(atface(bb.ymin, bb.ymax, ip[1][1]) && atface(bb.zmin, bb.zmax, ip[1][2]));
        atfaces.push_back(atface(bb.xmin, bb.xmax, ip[2][0]) && atface(bb.zmin, bb.zmax, ip[2][2]));
        atfaces.push_back(atface(bb.xmin, bb.xmax, ip[3][0]) && atface(bb.zmin, bb.zmax, ip[3][2]));
        atfaces.push_back(atface(bb.xmin, bb.xmax, ip[4][0]) && atface(bb.ymin, bb.ymax, ip[4][1]));
        atfaces.push_back(atface(bb.xmin, bb.xmax, ip[5][0]) && atface(bb.ymin, bb.ymax, ip[5][1]));
        std::vector<hitpoint> hitpoints;
        hitpoint tmphp;
        for(int i = 0; i<6; i++)
        {
            if(atfaces[i]){
                tmphp.t = t[i];
                tmphp.posx = ip[i][0];
                tmphp.posy = ip[i][1];
                tmphp.posz = ip[i][2];
                hitpoints.push_back(tmphp);
            }
        }
        
        if(hitpoints.size()>0)  // This should almost always be of length 2 if it hits the bb; unless it exactly hits the corner/edge of the bounding box
        {
            std::sort(hitpoints.begin(), hitpoints.end(), comparehitpointbytime);   // Sort by arrival time and the smallest will give us the face it hits first.
            out.hit = true;
            out.inside = false;
            out.x = hitpoints[0].posx;
            out.y = hitpoints[0].posy;
            out.z = hitpoints[0].posz;
            // Should this update tstart and tend? not sure…
            return(out);
        }
        else        // Does not hit the bounding box, mark the ray as such
        {
            out.hit = false;
            out.inside = false;
            return(out);
        }
    }
}

// Given a set of rays find the point which is closest to these lines (minimizing their square distance)
candidatematch ClosestPointToLines(std::map<std::pair<int, int>,transformedray>& raydb, std::vector<camrayid> crids)
{
    candidatematch out;
    out.camrayids = crids;
    long len = crids.size();
    double rhs[3];
    double lhs[3][3];
    double lhsnew[3][3];
    
    for(int i=0;i<3;i++)            // Identity matrix and zero rhs / Fill in matrix for each 3 dimensions
    {
        rhs[i] = 0.0;
        lhs[i][i] = len;
    }
    
    for(int i=0;i<len;i++)            // Fill in rhs and lhs for each ray
    {
        transformedray r;
        r = raydb[std::make_pair(crids[i].camid,crids[i].rayid)];
        
        lhs[0][0] -= r.vx*r.vx;
        lhs[1][0] -= r.vx*r.vy;
        lhs[2][0] -= r.vx*r.vz;
        lhs[0][1] -= r.vy*r.vx;
        lhs[1][1] -= r.vy*r.vy;
        lhs[2][1] -= r.vy*r.vz;
        lhs[0][2] -= r.vz*r.vx;
        lhs[1][2] -= r.vz*r.vy;
        lhs[2][2] -= r.vz*r.vz;
        
        double adotd = r.x*r.vx + r.y*r.vy + r.z*r.vz;
        rhs[0] += r.x - r.vx*adotd;
        rhs[1] += r.y - r.vy*adotd;
        rhs[2] += r.z - r.vz*adotd;
    }

    // Solve it
    double idet = -(lhs[0][2]*lhs[1][1]*lhs[2][0]) + lhs[0][1]*lhs[1][2]*lhs[2][0] + lhs[0][2]*lhs[1][0]*lhs[2][1] - lhs[0][0]*lhs[1][2]*lhs[2][1] - lhs[0][1]*lhs[1][0]*lhs[2][2] + lhs[0][0]*lhs[1][1]*lhs[2][2];
    
    idet = 1/idet;  // 1/determinant of LHS
    lhsnew[0][0] = -lhs[1][2]*lhs[2][1] + lhs[1][1]*lhs[2][2];
    lhsnew[0][1] =  lhs[0][2]*lhs[2][1] - lhs[0][1]*lhs[2][2];
    lhsnew[0][2] = -lhs[0][2]*lhs[1][1] + lhs[0][1]*lhs[1][2];
    lhsnew[1][0] =  lhs[1][2]*lhs[2][0] - lhs[1][0]*lhs[2][2];
    lhsnew[1][1] = -lhs[0][2]*lhs[2][0] + lhs[0][0]*lhs[2][2];
    lhsnew[1][2] =  lhs[0][2]*lhs[1][0] - lhs[0][0]*lhs[1][2];
    lhsnew[2][0] = -lhs[1][1]*lhs[2][0] + lhs[1][0]*lhs[2][1];
    lhsnew[2][1] =  lhs[0][1]*lhs[2][0] - lhs[0][0]*lhs[2][1];
    lhsnew[2][2] = -lhs[0][1]*lhs[1][0] + lhs[0][0]*lhs[1][1];
    for(int i = 0; i <3; i++)
    {
       for(int j = 0; j <3; j++)
        {
            lhs[i][j] = lhsnew[i][j] * idet;  // normally: divide cofactor by determinant, now multiplication because 1/determinant
        }
    }

    double sol[3];
    
    // sol = lhs*rhs;
    sol[0] = lhs[0][0] * rhs[0] + lhs[0][1] * rhs[1] + lhs[0][2] * rhs[2];
    sol[1] = lhs[1][0] * rhs[0] + lhs[1][1] * rhs[1] + lhs[1][2] * rhs[2];
    sol[2] = lhs[2][0] * rhs[0] + lhs[2][1] * rhs[1] + lhs[2][2] * rhs[2];
    
    //std::cout << sol[0] << std::endl;
    double error = 0;
    double solminusax, solminusay, solminusaz;
    double tmpx,tmpy,tmpz;
    for(int i=0;i<len;i++)            // calculate distances
    {
        transformedray r;
        r = raydb[std::make_pair(crids[i].camid,crids[i].rayid)];
        
        solminusax = sol[0] - r.x;
        solminusay = sol[1] - r.y;
        solminusaz = sol[2] - r.z;
        
        tmpx = r.vz * solminusay - r.vy * solminusaz;
        tmpy = solminusaz * r.vx - solminusax * r.vz;
        tmpz = solminusax * r.vy - solminusay * r.vx;
        
        error += tmpx*tmpx + tmpy*tmpy + tmpz*tmpz;
    }
    error = sqrt(error/len);
    
    out.matchx = sol[0];
    out.matchy = sol[1];
    out.matchz = sol[2];
    out.matcherror = error;
    return(out);
}

// Given a set of rays, bounds and other criteria find the matches
std::vector<candidatematch> SpaceTraversalMatching(const std::vector<ray>& raydata, const boundingboxspec& bb, const std::vector<std::vector<double>>& bounds, const int maxmatchesperray, const int& mincameras, const double& maxdistance, const double& deltat)
{
    // Prepare the rays
    std::map<std::pair<int, int>,transformedray> raydb;         // Store a dictionary of (cameraid, rayid): transformedray
    std::vector<transformedray> validrays;                      // Store the transformed rays
    validrays.reserve(raydata.size());
    std::map<int,int> numrays;                                  // Store the number of rays per camera
    std::map<int,int> invalidcounter;                           // Store the number that are invalid
    for(int i = 0; i < raydata.size(); i++)
    {
        if(numrays.find(raydata[i].camid) == numrays.end())     // Key does not yet exist  // maintain counter
            numrays.insert(std::make_pair(raydata[i].camid,1));
        else // Key exists
            numrays[raydata[i].camid]++;
        
        transformedray freshray;
        freshray = PrepareRay(raydata[i], bb, deltat);
        if(freshray.hit)
        {
            raydb.insert(std::make_pair(std::make_pair(freshray.camid, freshray.rayid),freshray));
            validrays.push_back(freshray);
        }
        else
        {
            if(invalidcounter.find(freshray.camid) == invalidcounter.end()) // Key does not yet exist  // maintain counter
                invalidcounter.insert(std::make_pair(freshray.camid,1));
            else // Key exists
                invalidcounter[freshray.camid]++;
        }
    }
    validrays.shrink_to_fit();  // Remove any access memory if possible

    // Print for each camera the number of rays
    std::cout << "# of rays for each camera: {";
    logstream << "# of rays for each camera: {";
    for (auto kv = numrays.begin(); kv != numrays.end();) {
        std::cout << kv->first << ": " << kv->second;
        logstream << kv->first << ": " << kv->second;
        if(++kv != numrays.end())
        {
            std::cout << ", ";
            logstream << ", ";
        }
    }
    std::cout << "}" << std::endl;
    logstream << "}" << std::endl;
  
    // Print for each camera the number of rays that miss the bounding box
    std::cout << "# of rays that miss the bounding box: {";
    logstream << "# of rays that miss the bounding box: {";
    for (auto kv = invalidcounter.begin(); kv != invalidcounter.end();) {
        std::cout << kv->first << ": " << kv->second;
        logstream << kv->first << ": " << kv->second;
        if(++kv != invalidcounter.end())
        {
            std::cout << ", ";
            logstream << ", ";
        }
    }
    std::cout << "}" << std::endl;
    logstream << "}" << std::endl;
    
    // print rays that miss the bounding box
//    for (auto kv : invalidcounter) {
//        std::cout << "Camera " << kv.first << " has " << kv.second << " rays that miss the bounding box" << std::endl;
//    }
    
    t2b = clock();
    // Do the actual traversing
    std::vector<traversedcell> traversed;
    std::vector<traversedcell> tmptraverse;
    std::vector<traversedcell> tmptraverse2;
    transformedray revvr;
    for (transformedray& vr : validrays)
    {
        tmptraverse = DirectionalVoxelTraversal(vr,bounds, vr.tend);         // Traverse forwards
        if(vr.inside)                                               // Ray is inside, traverse also backwards
        {
            revvr = vr;
            revvr.vx *= -1;
            revvr.vy *= -1;
            revvr.vz *= -1;
            tmptraverse2 = DirectionalVoxelTraversal(revvr,bounds, -vr.tstart); // Combine these two…
            tmptraverse.insert(tmptraverse.end(), tmptraverse2.begin(), tmptraverse2.end()); //
        }
        std::sort(tmptraverse.begin(),tmptraverse.end(),comparetraversedcell);
        tmptraverse.erase(std::unique(tmptraverse.begin(),tmptraverse.end(),comparetraversedcellsamecell),tmptraverse.end());
        //std::cout << tmptraverse.size() << " ";
        traversed.insert(traversed.end(), tmptraverse.begin(), tmptraverse.end());
    }
    t2e = clock();
    timing2 += double(t2e-t2b);
    
//    for(auto i: traversed)
//    {
//            std::cout << i.camid << "." << i.rayid << "  " << i.cellid.xi << "," << i.cellid.yi << "," << i.cellid.xi << std::endl;
//    }
    
    std::cout << "# of voxels traversed after expansion: " << traversed.size() << std::endl;
    logstream << "# of voxels traversed after expansion: " << traversed.size() << std::endl;
    
    t3b = clock();
    std::sort(traversed.begin(),traversed.end(),comparetraversedcell);  // Sort the traversed cells
    t3e = clock();
    timing3 += double(t3e-t3b);
    
    // Group the traversed cell (voxel) by cell ID, now these should be consecutive in the traversed vector
    std::vector<groupedcell> groupedcells;
    groupedcell tmpgroupedcell;
    tmpgroupedcell.camrayids.push_back({traversed[0].camid, traversed[0].rayid});
    tmpgroupedcell.cellid = traversed[0].cellid;
    for(long i = 1; i < traversed.size(); i++)
    {
        if(tmpgroupedcell.cellid.xi == traversed[i].cellid.xi && tmpgroupedcell.cellid.yi == traversed[i].cellid.yi && tmpgroupedcell.cellid.zi == traversed[i].cellid.zi)
        {
            tmpgroupedcell.camrayids.push_back({traversed[i].camid, traversed[i].rayid});
        }
        else
        {
            if(tmpgroupedcell.camrayids.size() >= mincameras)       // Rough prune number of rays should be bigger than or equal to mincameras, but could still be rays from the same camera!
            {
                std::vector<camrayid> tmpcamrayids = tmpgroupedcell.camrayids;
                std::sort(tmpcamrayids.begin(), tmpcamrayids.end(), comparecamrayidscam);
                long uniquecount = std::unique(tmpcamrayids.begin(), tmpcamrayids.end(), comparecamrayidsamecam) - tmpcamrayids.begin();
                if(uniquecount >= mincameras)                       // More careful prune
                {
                    groupedcells.push_back(tmpgroupedcell);
                }
            }
            tmpgroupedcell.cellid = traversed[i].cellid;
            tmpgroupedcell.camrayids.clear();
            tmpgroupedcell.camrayids.push_back({traversed[i].camid, traversed[i].rayid});
        }
    }
    // Process remainder:
    if(tmpgroupedcell.camrayids.size() >= mincameras)       // Rough prune
    {
        std::vector<camrayid> tmpcamrayids = tmpgroupedcell.camrayids;
        std::sort(tmpcamrayids.begin(), tmpcamrayids.end(), comparecamrayidscam);
        long uniquecount = std::unique(tmpcamrayids.begin(), tmpcamrayids.end(), comparecamrayidsamecam) - tmpcamrayids.begin();
        if(uniquecount >= mincameras)                       // More careful prune
        {
            groupedcells.push_back(tmpgroupedcell);
        }
    }
    std::cout << "Prune based on number of cameras: " << groupedcells.size() << std::endl;
    logstream << "Prune based on number of cameras: " << groupedcells.size() << std::endl;
    
     
    
    
    // Get rid of cellID only the groups of camid/rayids matter
    std::vector<std::vector<camrayid>> candidatepairs;
    candidatepairs.reserve(groupedcells.size());
    std::vector<camrayid> tmpcamrayids;
    for(const groupedcell& i: groupedcells)
    {
        tmpcamrayids = i.camrayids;
        std::sort(tmpcamrayids.begin(),tmpcamrayids.end(),comparecamrayidscamray);
        candidatepairs.push_back(tmpcamrayids);
    }
    
    t5b = clock();
    // This most of the time semi-sorted, so we can unique first and remove a lot…
    candidatepairs.erase(std::unique(candidatepairs.begin(), candidatepairs.end(), comparecamrayidsidentical), candidatepairs.end());
    std::sort(candidatepairs.begin(),candidatepairs.end(),comparecamrayidsordered);         // Now sort followed by another unique…
    candidatepairs.erase(std::unique(candidatepairs.begin(), candidatepairs.end(), comparecamrayidsidentical), candidatepairs.end());
    candidatepairs.shrink_to_fit();
    t5e = clock();
    timing5 += double(t5e-t5b);
   
    // From all these groupings ("pairs") find the camera combinations:
    t8b = clock();
    std::vector<std::vector<camrayid>> candidates;
    for(const std::vector<camrayid>& cand: candidatepairs)
    {
        // Split candidates in to groups based on camera:
        std::vector<std::vector<camrayid>> groupedcandidates;   //(cam 1, 1, 1, 2, 2, 3, 4, 4, 4 => [1,1,1],[2,2],[3],[4,4,4])
        std::vector<camrayid> tmpgroupedcandidates;
        tmpgroupedcandidates.push_back(cand[0]);
        for(long i = 1; i < cand.size(); i++)
        {
            
            if(tmpgroupedcandidates[0].camid == cand[i].camid)
            {
                tmpgroupedcandidates.push_back(cand[i]);
            }
            else
            {
                groupedcandidates.push_back(tmpgroupedcandidates);
                tmpgroupedcandidates.clear();
                tmpgroupedcandidates.push_back(cand[i]);
            }
        }
        groupedcandidates.push_back(tmpgroupedcandidates);
        
        
        std::vector<camrayid> tmp;
        std::vector<std::vector<camrayid>> newcandidates;
        GenerateCamRayIDPermutations(groupedcandidates, newcandidates, 0, tmp);
        candidates.insert(candidates.end(), newcandidates.begin(), newcandidates.end());
    }
//    std::cout << "Candidates size: " << candidates.size() << std::endl;
    t8e = clock();
    timing8 += double(t8e-t8b);
    
    // Sort all these camid/rayid groups and remove duplicates
    std::sort(candidates.begin(),candidates.end(),comparecamrayidsordered);
    candidates.erase(std::unique(candidates.begin(), candidates.end(), comparecamrayidsidentical), candidates.end());
   
    std::cout << "Duplicate candidates removed: " << candidates.size() << std::endl;
    logstream << "Duplicate candidates removed: " << candidates.size() << std::endl;
    candidates.shrink_to_fit(); // Free memory if possible
//    for(auto i: candidates)
//    {
//        std::cout << "candidates: ";
//        for (auto j: i)
//        {
//            std::cout << j.camid << "." <<j .rayid << " ";
//        }
//        std::cout << std::endl;
//    }
    
    // Calculate match positions and errors for each of the candidates
    std::vector<candidatematch> candidatematches;
    candidatematches.reserve(candidates.size());
    std::cout << "We have " << candidates.size() << " candidates" << std::endl;
    for(const std::vector<camrayid>& cand: candidates)
    {
        candidatematch tmp;
        tmp = ClosestPointToLines(raydb,cand);
        candidatematches.push_back(tmp);
    }
    
    std::sort(candidatematches.begin(), candidatematches.end(), comparecandidatematches); // Sort matches by number of cameras decreasing, and then match error increasing.
  
//    for(auto i: candidatematches)
//    {
//        std::cout << "Candidates match: ";
//        for (auto j: i.camrayids)
//        {
//            std::cout << j.camid << "." <<j .rayid << "\t";
//        }
//        std::cout << "\t| "  << i.matcherror << "\t| " << i.matchx << "\t"  << i.matchy << "\t"  << i.matchz << "\t" <<  std::endl;
////        for (auto j: i.camrayids)
////        {
////            transformedray r;
////            r = raydb[std::make_pair(j.camid,j.rayid)];
////            std::cout << r.x << "\t" << r.y << "\t" << r.z << "\t"  << r.vx << "\t" << r.vy << "\t" << r.vz << "\t" << std::endl;
////        }
////        std::cout << std::endl;
//    }
    
    // Select best matches from candidates
    std::cout << "Selecting the best matches with upto " << maxmatchesperray << " match(es)/ray out of " << candidates.size() <<" candidates" << std::endl;
    logstream << "Selecting the best matches with upto " << maxmatchesperray << " match(es)/ray out of " << candidates.size() <<" candidates" << std::endl;
    std::vector<candidatematch> approvedmatches;
    std::map<std::pair<long, long>,int> matchcounter;
    approvedmatches.clear();
    matchcounter.clear();
    bool valid;
    int matches;
    for(const candidatematch& cand: candidatematches)
    {
        if(cand.matcherror < maxdistance)
        {
            valid = true;
            for(const camrayid& camrayid: cand.camrayids)
            {
                std::pair<long, long> idpair;
                idpair = std::make_pair(camrayid.camid,camrayid.rayid);
                if(!(matchcounter.find(idpair) == matchcounter.end())) // If it exists
                {
                    matches = matchcounter[idpair];
                    if(matches >= maxmatchesperray)
                    {
                        valid = false;
                        break;
                    }
                }
            }
            
            
            
            if(valid)
            {
                // Add to approved matches
                for(const camrayid& camrayid: cand.camrayids)  // Add match counter
                {
                    std::pair<long, long> idpair;
                    idpair = std::make_pair(camrayid.camid,camrayid.rayid);
                    if(matchcounter.find(idpair) == matchcounter.end())                     // If does not exists
                    {
                       matchcounter.insert(std::pair<std::pair<long,long>,int>(idpair,1));  // Add to counter
                    }
                    else
                    {
                        matchcounter[idpair]++;                                             // Increment counter
                    }
                }
                approvedmatches.push_back(cand);
            }
        }
    }

    
    std::cout << "Selecting done. " << approvedmatches.size() << " matched found (out of " << candidates.size() << " candidates)" << std::endl;
    logstream << "Selecting done. " << approvedmatches.size() << " matched found (out of " << candidates.size() << " candidates)" << std::endl;
    
    /*
    int iter = 0;
    for(auto i: approvedmatches)
    {
        if(iter++<30000)
        {
            std::cout << "Approved match: ";
            for (auto j: i.camrayids)
            {
                std::cout << j.camid << "." <<j .rayid << "\t";
            }
            std::cout << "\t| "  << i.matcherror << std::endl;
        }
        else
        {
            break;
        }
    }
    */
    
    return(approvedmatches);
}

// Initializations setup
void init(){
    neighbours.clear();                 // Presumably it is already empty…
    neighbours.push_back({-1,0,0});     // Add neighbours that are above/below/left/right/front/back
    neighbours.push_back({0,-1,0});
    neighbours.push_back({0,0,-1});
    neighbours.push_back({0,0,1});
    neighbours.push_back({0,1,0});
    neighbours.push_back({1,0,0});
}

// Make a nice date-time string
std::string timeasstring()
{
    time_t rawtime;
    struct tm * timeinfo;
    char buffer[80];
    
    time (&rawtime);
    timeinfo = localtime(&rawtime);
    
    strftime(buffer,sizeof(buffer),"%d-%m-%Y %I:%M:%S",timeinfo);
    std::string str(buffer);
    
    return(str);
}

// Start of the program
int main(int argc, const char * argv[]) {
    std::clock_t tstart = clock();  // Start the clock
    
    std::cout << "O hai! Howz u?" << std::endl;
    std::cout << "Space-traversal-matching V0.2 (2020-11-24)" << std::endl << std::endl;
    
    init();
    
    switch(argc-1){  // Check the number of arguments (-1 because the program itself is included in argc)
        case 14:{                            // We have 14 arguments: 1:input file, 2:xmin, 3:xmax, 4:divx, 5:ymin, 6:ymax, 7:divy, 8:zmin, 9:zmax, 10:divz, 11:ncam, 12:match/ray, 13:maxerr, 14: delta
            unsigned int maxframes = 999999999;                 // Limit the number of frames to be processed can be set up to ~4 billion
            int mincameras = std::atoi(argv[11]);        // We require a match to have rays from at least mincameras of different cameras
            int maxmatchesperray = std::atoi(argv[12]);  // Maximum number of matches for each ray
            double maxdistance = std::atof(argv[13]);    // Maximum distance allowed for a match.
            
            double deltat = std::atof(argv[14]);         // Extend rays by this amount in both directions
            
            std::cout << "Matches at most " << maxframes << " frames" << std::endl;
            std::cout << "Minimum number of cameras for each match " << mincameras << std::endl;
            std::cout << "Each ray can be matched at most " << maxmatchesperray << " times" << std::endl;
            std::cout << "Max mean distance error for each match " << maxdistance << std::endl;
            
            int divisionsx = std::atoi(argv[4]);
            int divisionsy = std::atoi(argv[7]);
            int divisionsz = std::atoi(argv[10]);
            
            boundingboxspec bb;                       // Store bounding box data in bb
            
            bb.xmin = std::atof(argv[2]);             // Bounding box min and max for each direction
            bb.xmax = std::atof(argv[3]);
            bb.ymin = std::atof(argv[5]);
            bb.ymax = std::atof(argv[6]);
            bb.zmin = std::atof(argv[8]);
            bb.zmax = std::atof(argv[9]);
            
            std::cout << "x-range: " << bb.xmin << "  -  " << bb.xmax << " with " << divisionsx << " divisions" << std::endl;
            std::cout << "y-range: " << bb.ymin << "  -  " << bb.ymax << " with " << divisionsy << " divisions" << std::endl;
            std::cout << "z-range: " << bb.zmin << "  -  " << bb.zmax << " with " << divisionsz << " divisions" << std::endl;

            bb.nx = divisionsx;                    // Number of divisions in each direction
            bb.ny = divisionsy;
            bb.nz = divisionsz;
            
            
            // Open all the input/output streams logstream is a global variable.
            std::string inputfile = argv[1];
            std::string filebasename = inputfile;
            size_t lastindex = filebasename.find_last_of(".");
            filebasename = filebasename.substr(0, lastindex);
            std::string logfile = filebasename;
            std::string outputfile = filebasename;
            outputfile.append("_out.bin");
            logfile.append("_out.log");
            std::cout << "Input file:  " << inputfile << std::endl;
            std::cout << "Output file: " << outputfile << std::endl;
            std::cout << "Log file:    " << logfile << std::endl;
            
            std::ofstream streamout;
            streamout.open(outputfile, std::ios::out | std::ios::binary);
            logstream.open(logfile,std::ios::out);
            
            std::ifstream rayfile;
            rayfile.open (inputfile, std::ios::in | std::ios::binary);

            
            std::vector<std::vector<double>> bounds;    // Initialize bounds vectors, from the bb specification find all the boundaries of all the voxels based on the number of divisions.
            std::vector<double> tmp;
            tmp.reserve(bb.nx);
            tmp.clear();
            for(int i = 0; i<bb.nx;i++)
            {
                tmp.push_back(bb.xmin + i*(bb.xmax-bb.xmin)/bb.nx);
            }
            tmp.push_back(bb.xmax);
            bounds.push_back(tmp);
            
            tmp.clear();
            tmp.reserve(bb.ny);
            for(int i = 0; i<bb.ny;i++)
            {
                tmp.push_back(bb.ymin + i*(bb.ymax-bb.ymin)/bb.ny);
            }
            tmp.push_back(bb.ymax);
            bounds.push_back(tmp);
            
            tmp.clear();
            tmp.reserve(bb.nz);
            for(int i = 0; i<bb.nz;i++)
            {
                tmp.push_back(bb.zmin + i*(bb.zmax-bb.zmin)/bb.nz);
                //std::cout << tmp[tmp.size()-1] << " ";
            }
            tmp.push_back(bb.zmax);
            //std::cout << tmp[tmp.size()-1] << " ";
            bounds.push_back(tmp);
            
            
            std::cout << "# of cells: " << bb.nx*bb.ny*bb.nz << std::endl;
            logstream << "# of cells: " << bb.nx*bb.ny*bb.nz << std::endl;
            
            // Start reading in data
            int currentframe = 0;
            if (rayfile.is_open()) {
                std::cout << argv[1] << " opened!" << std::endl << std::endl;
                logstream << argv[1] << " opened!" << std::endl << std::endl;
                
                uint32_t numrays = 0;
                rayfile.read((char*)&numrays,sizeof(numrays));      // Read in number of rays for the first frame
                currentframe++;
                
                // Loop over each frame either till the end of the file or if maxframes is reached
                while(!rayfile.eof() && currentframe <= maxframes)
                {
                    std::cout << "####### Frame (rays): " << currentframe << "(" << numrays << ")" << std::endl;
                    logstream << "####### Frame (rays): " << currentframe << "(" << numrays << ")" << std::endl;
                    auto timestring = timeasstring();
                    std::cout << timestring << std::endl;
                    logstream << timestring << std::endl;
                    // Read array of rays
                    std::vector<ray> rays;
                    int tmp = 0;
                    for(int i =0; i < numrays; i++)     // Read in numrays of rays
                    {
                        tmp++;
                        ray tmpray;
                        uint8_t tmpcamid;
                        uint16_t tmprayid;
                        float tmpx,tmpy,tmpz,tmpvx,tmpvy,tmpvz, rayt1, rayt2;  // Floats (4 byte)
                        
                        rayfile.read((char*)&tmpcamid,sizeof(tmpcamid));        // Read in Cam ID, Ray ID, x, y, z, vx, vy, vz, tstart, tend
                        rayfile.read((char*)&tmprayid,sizeof(tmprayid));
                        rayfile.read((char*)&tmpx,sizeof(tmpx));
                        rayfile.read((char*)&tmpy,sizeof(tmpy));
                        rayfile.read((char*)&tmpz,sizeof(tmpz));
                        rayfile.read((char*)&tmpvx,sizeof(tmpvx));
                        rayfile.read((char*)&tmpvy,sizeof(tmpvy));
                        rayfile.read((char*)&tmpvz,sizeof(tmpvz));
                        rayfile.read((char*)&rayt1,sizeof(rayt1));
                        rayfile.read((char*)&rayt2,sizeof(rayt2));
                        tmpray.camid = tmpcamid;
                        tmpray.rayid = tmprayid;
                        tmpray.x = tmpx;
                        tmpray.y = tmpy;
                        tmpray.z = tmpz;
                        tmpray.vx = tmpvx;
                        tmpray.vy = tmpvy;
                        tmpray.vz = tmpvz;
                        tmpray.tstart = rayt1;
                        tmpray.tend = rayt2;
                        
                        rays.push_back(tmpray);
                    }
                    
                    std::vector<candidatematch> results;
            
                    
                    // Do the actual matching, function returns the matches
                    results = SpaceTraversalMatching(rays, bb, bounds, maxmatchesperray, mincameras, maxdistance, deltat);
                    
                    // Write matches to file
                    uint32_t numberofmatches = (int)results.size();
                    streamout.write((char*)&numberofmatches, sizeof(uint32_t));
                    std::cout << "Number of matches: " << numberofmatches << std::endl;
                    logstream << "Number of matches: " << numberofmatches << std::endl;
                    for(const candidatematch& match: results)
                    {
                        uint8_t numberofcams = (int)match.camrayids.size();     // Write number of cams/rays that were used for the match
                        streamout.write((char*)&numberofcams, sizeof(uint8_t));
                        
                        float val = match.matchx;                                 // Write x, y, z, and match error to file
                        streamout.write((char*)&val, sizeof(float));
                        val = match.matchy;
                        streamout.write((char*)&val, sizeof(float));
                        val = match.matchz;
                        streamout.write((char*)&val, sizeof(float));
                        val = match.matcherror;
                        streamout.write((char*)&val, sizeof(float));
                        for(const camrayid& camrayid:match.camrayids)       // Write cam id-ray id pairs:
                        {
                            uint8_t camid = (int)camrayid.camid;
                            uint16_t rayid =(int)camrayid.rayid;
                            streamout.write((char*)&camid, sizeof(uint8_t));
                            streamout.write((char*)&rayid, sizeof(uint16_t));
                        }
                    }
                    
                    rayfile.read((char*)&numrays,sizeof(numrays));  // Read number of rays for next frame
                    currentframe++;
                }
            }
            else{
                std::cout << "Something went wrong during opening the file. Does the file exist?" << std::endl;
                logstream << "Something went wrong during opening the file. Does the file exist?" << std::endl;
            }
            rayfile.close();                // Close ray file
            
            
            std::clock_t tend = clock();
            double timing = double(tend - tstart) / CLOCKS_PER_SEC;
            std::cout  << std::endl << "Elapsed time = " << timing << " sec" << std::endl;
            logstream  << std::endl << "Elapsed time = " << timing << " sec" << std::endl;
            timing /= currentframe;
            std::cout  << std::endl << "Elapsed time/frame = " << timing << " sec" << std::endl;
            logstream  << std::endl << "Elapsed time/frame = " << timing << " sec" << std::endl;
            
            
            std::cout << "Partial timings:" << std::endl;
            std::cout << "1: " << timing1/maxframes << " clocks" << std::endl;
            std::cout << "2: " << timing2/maxframes << " clocks" << std::endl;
            std::cout << "3: " << timing3/maxframes << " clocks" << std::endl;
            std::cout << "4: " << timing4/maxframes << " clocks" << std::endl;
            std::cout << "5: " << timing5/maxframes << " clocks" << std::endl;
            std::cout << "6: " << timing6/maxframes << " clocks" << std::endl;
            std::cout << "7: " << timing7/maxframes << " clocks" << std::endl;
            std::cout << "8: " << timing8/maxframes << " clocks" << std::endl;
            std::cout << "9: " << timing9/maxframes << " clocks" << std::endl;
            std::cout << "10:" << timing10/maxframes << " clocks" << std::endl;
            

            streamout.close();      // Close output files (log and matches)
            logstream.close();
            
            std::cout << std::endl << "kthnxbye" << std::endl;
            logstream << std::endl << "kthnxbye" << std::endl;
        }
        break;
        default:{
            std::cout << (argc-1) << " arguments given where 14 were expected!" << std::endl; // first argument is the program itself
            std::cout << "<input file> <xmin> <xmax> <divx> <ymin> <ymax> <divy> <zmin> <zmax> <divz> <ncam> <max matches per ray> <max match error> <delta>" << std::endl;
        }
    }
    return 0;
}
